/* $Id$ */
/*
 * Copyright (c) 2003-2004 Jose Nazario <jose@monkey.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Jose Nazario.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/param.h>
#include <sys/time.h>
#include <sys/types.h>

#include <netinet/in.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <event.h>
#include <dnet.h>
#include <pcap.h>

#include "jscan.h"
#include "pcaputil.h"

void
usage(char *progname)
{
	fprintf(stderr, "usage: %s -t type [-i intf] [-d dst] [-f osfile] [-s src]\n",
		progname);
	fprintf(stderr, "\twhere type is active or passive\n");
	exit(0);
}

/* stolen from openbsd libc, needed my own type */

char           *
inet_ntoasc(ip_addr_t in)
{
	static char     b[18];
	register char  *p;

	p = (char *) &in;
#define UC(b)        (((int)b)&0xff)
	(void) snprintf(b, sizeof(b),
		     "%u.%u.%u.%u", UC(p[0]), UC(p[1]), UC(p[2]), UC(p[3]));
	return (b);
}

/*
 * fingerprint the OS using the SYN-ACK's TCP window size, IP TTL and DF
 * bit. common routines, basically stolen from siphon 0.666.
 */

char           *
osprint(struct myctx * ctx, uint16_t win, uint8_t ttl, uint16_t off,
	u_int16_t syn)
{
	FILE           *osprints;
	static char     line[80];
	static char     os[70] = "unknown";
	int             df = 0;	/* DF bit */
	int             match = 0;

	osprints = fopen(ctx->osfile, "r");
	if (!osprints)
		return (NULL);

	/* look for the DF bit, off is already masked by 0x1fff */
	if (off & 0x40)
		df = 1;

	/* cycle through the pf.os file, looking for prints */
	while (1) {
		static char    *oswin, *osttl, *osdf, *ossyn, *osos, *ostmp;

		fgets(line, 80, osprints);
		if (feof(osprints))
			break;

		/*
		 * pf.os line format:
		 * wwww:ttt:D:ss:OOO...:OS:Version:Subtype:Details
		 * 65535:64:1:60:M*,N,W1,N,N,T:FreeBSD:5.0-5.1::FreeBSD
		 * 4.7-5.1 from OpenBSD's PF. tokenize.
		 */
		oswin = strtok(line, ":");	/* window size */
		osttl = strtok(NULL, ":");	/* IP TTL */
		osdf = strtok(NULL, ":");	/* DF bit */
		ossyn = strtok(NULL, ":");	/* SYN size */
		ostmp = strtok(NULL, ":");	/* not used .. */
		ostmp = strtok(NULL, ":");	/* not used .. */
		ostmp = strtok(NULL, ":");	/* not used .. */
		osos = strtok(NULL, "\n");	/* OS line */

		/* valid tuples only ... */
		if (!(oswin) || !(osttl) || !(osdf) || !(ossyn))
			continue;

		if (strstr(oswin, "#") != NULL)
			continue;

		/*
		 * checks 1 and 2: seen window size == listed window size,
		 * and SYN-ACK size is what we expect.
		 */

		if ((atoi(ossyn) == syn) && ((win == atoi(oswin)) || (strstr(oswin, "*") != NULL))) {

			/* check 3: DF bit seen matches stated DF bit */

			if (df == atoi(osdf)) {

				/*
				 * check 4, the TTL checks. this assumes not
				 * so many hops are crossed that you obscure
				 * the original TTL, so we assume that it is
				 * the next common one rounded up. would be
				 * faster to mod() ...
				 */

				if (atoi(osttl) == 64 && ttl <= 64 && ttl > 32) {
					snprintf(os, sizeof(os), "%s", osos);
					match = 1;
				}
				if (atoi(osttl) == 255 && ttl <= 255 && ttl > 128) {
					snprintf(os, sizeof(os), "%s", osos);
					match = 1;
				}
				if (atoi(osttl) == 128 && ttl <= 128 && ttl > 64) {
					snprintf(os, sizeof(os), "%s", osos);
					match = 1;
				}
			}
		}
		if (match)
			break;
	}

	fclose(osprints);
	if (match)
		return (os);
	else
		return (NULL);
}


static void
report(u_char * packet, void *arg)
{
	struct myctx   *ctx = (struct myctx *) arg;
	static struct ip_hdr *ip_h;
	u_char         *tmp;
	const char     *p;
	struct addr     ip_src;
	static struct entry *np, *n2;

	tmp = packet + ctx->dl_len;
	ip_h = (struct ip_hdr *) tmp;
	if (ip_h->ip_v != 4)
		return;
	p = inet_ntoasc(ip_h->ip_src);
	if ((addr_aton(p, &ip_src)) == -1)
		return;

	/*
	 * if it's a passive scan, don't care about the src address. if it's
	 * an active scan, make sure it was the dest we specified. make sure
	 * it's a TCP packet, too, and has SA set.
	 */
	if (((ctx->flags == SCAN_FLAGS_PASSIVE) ||
	     ((addr_cmp(&ip_src, &(ctx->dst))) == 0)) &&
	    (ip_h->ip_p == IP_PROTO_TCP)) {
		struct tcp_hdr *tcp_h = (struct tcp_hdr *) (tmp + IP_HDR_LEN);

		if (tcp_h->th_flags == 0x12) {
			struct servent *serv;
			char           *s_name = "unknown", *os = NULL;

			TAILQ_FOREACH(np, &head, entries) {
				if ((addr_cmp(&ip_src, &(np->host)) == 0) &&
				    (np->port == tcp_h->th_sport))
					return;	/* seen */
			}

			if (ctx->osfile != NULL) {
				/* check the OS from the database */
				os = osprint(ctx, ntohs(tcp_h->th_win),
					     ip_h->ip_ttl, ip_h->ip_off,
					     ntohs(ip_h->ip_len));
				if (os == NULL)
					os = "unknown";
			}
			/* get the name of the service */
			serv = getservbyport(tcp_h->th_sport, "tcp");
			if (serv != NULL)
				s_name = strdup(serv->s_name);

			/* print it out */
			printf("%-16s %35s %15s %6d/tcp\n", addr_ntoa(&ip_src),
			       os ? os : "", s_name, htons(tcp_h->th_sport));

			/* new entry to the "seen" list */
			n2 = malloc(sizeof(struct entry));
			if (n2 == NULL)
				return;	/* XXX - but don't die */
			addr_aton(p, &(n2->host));
			n2->port = tcp_h->th_sport;
			n2->seen = time(NULL);
			TAILQ_INSERT_TAIL(&head, n2, entries);
		}
	}
	fflush(stdout);
	return;
}

static void
_recv(int fd, short event, void *arg)
{
	struct myctx   *ctx = (struct myctx *) arg;
	struct pcap_pkthdr ph;
	u_char         *pread;

	if ((ctx->flags == SCAN_FLAGS_ACTIVE) && (ctx->dport > 65535));
	else
		event_add(&ctx->recv_ev, &ctx->tv);	/* reschedule */

	if ((pread = (u_char *) pcap_next(ctx->p, &ph)) != NULL)
		report(pread, ctx);

	return;
}

static void
_send(int fd, short event, void *arg)
{
	struct myctx   *ctx = (struct myctx *) arg;
	struct jscan_pkt *pkt;
	u_char          buf[BUFSIZ];
	int             len = IP_HDR_LEN + TCP_HDR_LEN, dport;

	if (ctx->dport > 65535)
		return;

	/* reschedule */
	event_add(&ctx->send_ev, &ctx->tv);

	pkt = (struct jscan_pkt *) buf;
	dport = ctx->dport;

	ip_pack_hdr(&pkt->pkt_hdr_i.ip, IP_TOS_LOWDELAY, len,
		    rand_uint16(ctx->rand), 0, 128, IP_PROTO_TCP,
		    ctx->src.addr_ip, ctx->dst.addr_ip);
	tcp_pack_hdr(&pkt->pkt_hdr_t.tcp, rand_uint16(ctx->rand), ctx->dport,
		     rand_uint32(ctx->rand), rand_uint32(ctx->rand),
		     TH_SYN, rand_uint16(ctx->rand), 0);

	ip_checksum(pkt, len);
	if (ip_send(ctx->ip, pkt, len) < 0)
		warn("ip_send in _send, short send");
	else
		ctx->dport += 1;/* we SYNed that port */
	return;
}

int
main(int argc, char *argv[])
{
	struct myctx    ctx;
	char            ebuff[PCAP_ERRBUF_SIZE];
	int             p_fd, ch, type = 0;
	char           *intf = "fxp0", *scan_type = "passive", *src, *dst;
	struct entry   *n1;
	time_t          start;

	ctx.flags = SCAN_FLAGS_PASSIVE;
	ctx.osfile = NULL;

	if (argc < 2)
		usage(argv[0]);

	while ((ch = getopt(argc, argv, "d:f:i:s:t:h")) != -1) {
		switch (ch) {
		case 'd':
			dst = optarg;
			if (addr_aton(dst, &ctx.dst) == -1)
				err(1, "malformed destination address.");
			break;
		case 'f':
			ctx.osfile = optarg;
			break;
		case 'i':
			intf = optarg;
			break;
		case 's':
			src = optarg;
			if (addr_aton(src, &ctx.src) == -1)
				err(1, "malformed source address.");
			break;
		case 't':
			if (strncmp(optarg, "passive", 4) == 0) {
				if (type)
					err(1, "already chose a scan type");
				ctx.flags = SCAN_FLAGS_PASSIVE;
				type = 1;
			}
			if (strncmp(optarg, "active", 3) == 0) {
				if (type)
					err(1, "already chose a scan type");
				ctx.flags = SCAN_FLAGS_ACTIVE;
				type = 1;
			}
			scan_type = optarg;
			break;
		default:
			usage(argv[0]);
			break;
		}
	}
	argc -= optind;
	argv += optind;

	ctx.p = pcap_open_live(intf, 1500, (ctx.flags == SCAN_FLAGS_PASSIVE),
			       500, ebuff);
	if (ctx.p == NULL)
		err(1, "pcap_open_live");

	ctx.dl_len = pcap_dloff(ctx.p);

	event_init();

	ctx.tv.tv_sec = 0;
	ctx.tv.tv_usec = 500;

	p_fd = pcap_fileno(ctx.p);
	event_set(&ctx.recv_ev, p_fd, EV_READ, _recv, (void *) &ctx);
	event_add(&ctx.recv_ev, &ctx.tv);
	if (ctx.flags == SCAN_FLAGS_ACTIVE) {
		ctx.ip = ip_open();
		if (ctx.ip == NULL)
			err(1, "ip_open() failed ..");
		event_set(&ctx.send_ev, p_fd, EV_WRITE, _send, (void *) &ctx);
		event_add(&ctx.send_ev, &ctx.tv);
		ctx.dport = 1;
	}
	/* initialize the "seen" list */
	TAILQ_INIT(&head);
	n1 = malloc(sizeof(struct entry));
	if (n1 == NULL)
		err(1, "initialization failure in malloc.");
	addr_aton("0.0.0.1", &(n1->host));
	n1->port = 0;
	n1->seen = time(NULL);
	TAILQ_INSERT_HEAD(&head, n1, entries);

	ctx.rand = rand_open();

	printf("scan started, type is %s, listening on %s\n",
	       scan_type, intf);

	start = time(NULL);
	event_dispatch();

	printf("scan completed. total execution time was %d seconds.\n",
	       time(NULL) - start);
	return (1);
}
